# -*- coding: utf-8 -*-

#  PycaWM
#  pycawm.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# 'with' stuff
from __future__ import with_statement

import sys

if sys.version_info < (2, 5):
    raise ImportError('pycawm module needs at least Python 2.5!')

import Image
import os

from atexit import register as atexit_register
from select import select
from threading import Lock
from traceback import print_exc
from Xlib import X, Xatom, Xutil, display, protocol, Xcursorfont, error

from atomsdb import atomsdb
from basedirspec import load_first_data
from client import Client
from decorators import managed_client
from hookmanager import (MetaHookable, run_hooks, add_pre_hook, add_post_hook,
                         remove_pre_hook, remove_post_hook)
from misc import (good_key, good_mask, find_numlock_mask,
                  good_grab_key, good_ungrab_key)
from wrappers import DisplayWrapper
import debug
import focuses
import placements

SOCK_NAME = '/tmp/pycawm-%s' % os.getuid()

#import warnings
#warnings.simplefilter('ignore', DeprecationWarning)

def cleanup_unix_socket():
    if os.path.exists(SOCK_NAME):
        os.remove(SOCK_NAME)

class PycaWM(object):
    __metaclass__ = MetaHookable
    # wm states
    idle, moving, resizing = range(3)
    # resizing edges
    left, top, right, bottom = range(3, 7)
    focuses = dict()
    placements = dict()

    def __init__(self, disp=None):
        self.edge = None,
        self.colors = dict()
        self.cursors = dict()
        self.clients = Client.clients
        self.keys = dict()
        self.dispatch = {
            X.ConfigureRequest: self.configure_request_reply,
            X.MapRequest: self.map_request_reply,
            X.UnmapNotify: self.unmap_reply,
            X.Expose: self.expose_reply,
            X.ButtonPress: self.button_press_reply,
            X.ButtonRelease: self.button_release_reply,
            X.KeyPress: self.key_press_reply,
            X.KeyRelease: self.key_release_reply,
            X.PropertyNotify: self.property_reply,
            X.MotionNotify: self.motion_reply,
            X.EnterNotify: self.enter_reply,
            X.LeaveNotify: self.leave_reply,
            X.GravityNotify: self.gravity_reply,
            X.DestroyNotify: self.destroy_reply,
            X.CirculateRequest: self.circulate_request_reply,
            X.ClientMessage: self.client_message_reply,
            }
        self.plugins = list()
        self.state = PycaWM.idle
        self.start_positions = [0, 0]
        if disp is None:
            disp = os.getenv('DISPLAY', default=':0.0')
        os.putenv('DISPLAY', disp)
        try:
            self.display = display.Display(disp)
        except:
            print 'PycaWM: cannot open display %s ' % disp
            raise SystemExit(-1)
        self.screen = self.display.screen()
        self.root = self.screen.root
        self.check_for_other_wm()
        # okay, we can now change stuff!
        # intern and cache the atoms
        atomsdb.build_db(self.display)
        # create all the cursors
        self.create_cursors()
        self.root.change_attributes(cursor=self.cursors[self.state,],
                                    event_mask=(X.ColormapChangeMask |
                                                X.PropertyChangeMask |
                                                X.PointerMotionMask |
                                                X.ButtonPressMask |
                                                X.EnterWindowMask |
                                                X.SubstructureRedirectMask |
                                                X.SubstructureNotifyMask))
        self.create_focuses()
        self.current_focus = None
        self.create_placements()
        self.preferred_placements = list()
        self.display.sync()
        self.displays = [DisplayWrapper(self.display, self.handle_events)]
        self.sockets = list()
        self.event_lock = Lock()
        find_numlock_mask(self.display)
        atexit_register(cleanup_unix_socket)
        if not self.display.has_extension('SHAPE'):
            print 'no SHAPE extension for this X server!'
        else:
            self.dispatch[self.display.extension_event.ShapeNotify] = \
                self.shape_reply

    def check_for_other_wm(self):
        handler = error.CatchError(error.BadAccess)
        self.root.change_attributes(event_mask=X.SubstructureRedirectMask,
                                    onerror=handler)
        self.display.sync()
        # 2 == X_ChangeWindowAttributes
        if handler.get_error() and handler.error.major_opcode == 2:
            print 'PycaWM: another window manager is running!'
            raise SystemExit(-1)

    def get_color(self, color):
        if color in self.colors:
            return self.colors[color]
        if isinstance(color, basestring):
            col = self.screen.default_colormap.alloc_named_color(color)
            if col:
                self.colors[color] = col.pixel
                return col.pixel
            if len(color) == 7 and color[0] == '#' or len(color) == 6:
                red = int('0x%s' % color[-6:-4], 0) * 257
                green = int('0x%s' % color[-4:-2], 0) * 257
                blue = int('0x%s' % color[-2:], 0) * 257
                col = self.screen.default_colormap.alloc_color(
                    red, green, blue).pixel
                self.colors[color] = col
                return col
        elif isinstance(color, (tuple, list)) and len(color == 3):
            for component in color:
                if component < 0 or component > 255:
                    raise Exception('no color %s' % color)
            col = self.screen.default_colormap.alloc_color(
                color[0] * 257, color[1] * 257, color[2] * 257).pixel
            self.colors[color] = col
            return col
        raise Exception('no color %s' % color)

    def free_colors(self):
        self.screen.default_colormap.free_colors(self.colors.values(), 0L)


    def create_cursors(self):
        def create_cursor(glyph):
            return cursor_font.create_glyph_cursor(cursor_font, glyph,
                                                   glyph + 1, (0, 0, 0),
                                                   (0, 0, 0xffff))
        cursor_font = self.display.open_font('cursor')
        self.cursors[None,] = create_cursor(Xcursorfont.top_left_arrow)
        self.cursors[PycaWM.idle,] = create_cursor(Xcursorfont.top_left_arrow)
        self.cursors[PycaWM.moving,] = create_cursor(Xcursorfont.fleur)
        self.cursors[PycaWM.left,] = create_cursor(Xcursorfont.left_side)
        self.cursors[PycaWM.top,] = create_cursor(Xcursorfont.top_side)
        self.cursors[PycaWM.right,] = create_cursor(Xcursorfont.right_side)
        self.cursors[PycaWM.bottom,] = create_cursor(Xcursorfont.bottom_side)
        self.cursors[PycaWM.top, PycaWM.left] = create_cursor(
            Xcursorfont.top_left_corner)
        self.cursors[PycaWM.top, PycaWM.right] = create_cursor(
            Xcursorfont.top_right_corner)
        self.cursors[PycaWM.bottom, PycaWM.right] = create_cursor(
            Xcursorfont.bottom_right_corner)
        self.cursors[PycaWM.bottom, PycaWM.left] = create_cursor(
            Xcursorfont.bottom_left_corner)
        cursor_font.close()

    def free_cursors(self):
        for cursor in self.cursors.values():
            cursor.free()

    def add_key(self, keys, fun):
        keysym, mask = good_key(keys)
        if keysym != X.NoSymbol and mask != 0:
            if (keysym, mask) not in self.keys:
                self.keys[(keysym, mask)] = fun
                good_grab_key(self.root,
                              self.display.keysym_to_keycode(keysym),
                              mask)
                self.display.sync()

    def remove_key(self, keys):
        keysym, mask = good_key(keys)
        if keysym != X.NoSymbol and mask != 0:
            if (keysym, mask) in self.keys:
                del self.keys[(keysym, mask)]
                good_ungrab_key(self.root,
                                self.display.keysym_to_keycode(keysym),
                                mask)
                self.display.sync()

    def set_wallpaper(self, wallpaper_path, scale=True):
        wallpaper_path = load_first_data(wallpaper_path)
        if not os.path.isfile(wallpaper_path):
            print 'the wallpaper %s does not exist!' % wallpaper_path
            return

        # TODO: add ESETROOT_PMAP_ID and _XROOTPMAP_ID support
        geo = self.root.get_geometry()
        pixmap = self.root.create_pixmap(geo.width, geo.height,
                                         self.screen.root_depth)
        wallpaper = Image.open(wallpaper_path)
        if scale:
            size = wallpaper.size
            if geo.width * geo.height > size[0] * size[1]:
                # bicubic is best for upsampling
                resize_filter = Image.BICUBIC
            else:
                # ... and antialias for downsampling
                resize_filter = Image.ANTIALIAS
            wallpaper = wallpaper.resize((geo.width, geo.height), resize_filter)
        gc = self.root.create_gc()
        pixmap.put_pil_image(gc, 0, 0, wallpaper)
        gc.free()
        self.root.change_attributes(background_pixmap=pixmap)
        pixmap.free()
        self.root.clear_area()
        self.display.sync()

    def add_plugin(self, plugin, params=dict()):
        self.plugins.append(plugin(self, **params))

    def get_managed_client(self, window):
        clients_found = [client for client in self.clients
                         if (client.window == window or
                             client.parent == window or
                             client.title_bar == window)]
        if len(clients_found) > 1:
            debug.log('a client is more than once in the clients list')
        try:
            client = clients_found.pop()
        except IndexError:
            client = None
        return client

    def is_client_managed(self, window):
        client = self.get_managed_client(window)
        return client is not None

    def add_client(self, window, parent):
        return Client(window, parent, self)

    def manage_existing_windows(self):
        client = None
        windows = self.root.query_tree().children
        unmanaged_windows = [window for window in windows
                             if (not self.is_client_managed(window) and
                                 window.get_wm_state() and
                                 not window.get_attributes().override_redirect)]
        for window in unmanaged_windows:
            client = self.add_client(window, self.root)
            client.reparent()
            client.map()
        if client is not None:
            self.give_focus(client)

    def unmanage_windows(self):
        for client in self.clients[:]:
            client.unmanage()

    def configure_request_reply(self, event):
        def create_change():
            change = dict()
            if event.value_mask & X.CWX:
                change['x'] = event.x
            if event.value_mask & X.CWY:
                change['y'] = event.y
            if event.value_mask & X.CWWidth:
                change['width'] = event.width
            if event.value_mask & X.CWHeight:
                change['height'] = event.height
            if event.value_mask & X.CWStackMode:
                change['stack_mode'] = event.stack_mode
            return change
        client = self.get_managed_client(event.window)
        if client is not None:
            client.update_size(create_change())
        else:
            # just accept the requested values
            event.window.configure(**create_change())

    def map_request_reply(self, event):
        client = self.get_managed_client(event.window)
        first_time = client is None
        if first_time and not event.window.get_attributes().override_redirect:
            client = self.add_client(event.window, self.root)
        if client is None:
            return
        if first_time:
            client.manage()
        else:
            client.map()

    @managed_client
    def unmap_reply(self, event, client):
        # consider only the UnmapNotifies from the real window
        if event.window == client.window:
            client.withdraw()

    @managed_client
    def expose_reply(self, event, client):
        # we only handle the last expose event
        if event.count != 0:
            return
        # redraw the window frame
        client.update_frame()

    def start_moving(self, client):
        # window moving
        self.state = PycaWM.moving
        self.display.change_active_pointer_grab(X.ButtonReleaseMask |
                                                X.ButtonMotionMask,
                                                self.cursors[self.state,],
                                                X.CurrentTime)
        pointer = self.root.query_pointer()
        self.start_positions = [pointer.root_x, pointer.root_y]

    def find_edge(self, client, x, y):
        edge = None,

        if x in range(client.parent_border_width):
            edge = PycaWM.left,
        elif y in range(client.parent_border_width):
            edge = PycaWM.top,
        elif x in range(
            client.geometry.width + client.parent_border_width,
            client.geometry.width + (client.parent_border_width * 2)):
            edge = PycaWM.right,
        elif y in range(
            client.geometry.height + client.parent_border_width,
            client.geometry.height + (client.parent_border_width * 2)):
            edge = PycaWM.bottom,

        if ((PycaWM.left in edge and
             y < client.title_bar_height + client.parent_border_width) or
            (PycaWM.top in edge and
             x < client.title_bar_height + client.parent_border_width)):
            edge = PycaWM.top, PycaWM.left
        elif ((PycaWM.right in edge and
               y < client.title_bar_height + client.parent_border_width) or
              (PycaWM.top in edge and
               x > client.geometry.width + client.parent_border_width -
               client.title_bar_height)):
            edge = PycaWM.top, PycaWM.right
        elif ((PycaWM.right in edge and
               y > client.geometry.height + client.parent_border_width -
               client.title_bar_height) or
              (PycaWM.bottom in edge and
               x > client.geometry.width + client.parent_border_width -
               client.title_bar_height)):
            edge = PycaWM.bottom, PycaWM.right
        elif ((PycaWM.left in edge and
               y > client.geometry.height + client.parent_border_width -
               client.title_bar_height) or
              (PycaWM.bottom in edge and
               x < client.title_bar_height + client.parent_border_width)):
            edge = PycaWM.bottom, PycaWM.left

        return edge

    def start_resizing(self, client):
        self.state = PycaWM.resizing
        pointer = self.root.query_pointer()
        x = pointer.root_x - client.geometry.x
        y = pointer.root_y - client.geometry.y
        self.edge = self.find_edge(client, x, y)

        self.display.change_active_pointer_grab(X.ButtonReleaseMask |
                                                X.ButtonMotionMask,
                                                self.cursors[self.edge],
                                                X.CurrentTime)
        self.start_positions = [pointer.root_x, pointer.root_y]

    @managed_client
    def button_press_reply(self, event, client):
        if self.state != PycaWM.idle:
            return
        if event.window == client.title_bar and \
                client.button_pressed(event.event_x, event.event_y) is None:
                self.start_moving(client)
        elif event.window == client.parent:
            self.start_resizing(client)

    def button_release_reply(self, event):
        client = self.get_managed_client(event.window)
        if client is not None and event.window == client.title_bar:
            client.button_press(event.event_x, event.event_y, event.detail)
        # reset the state and start_positions
        # (useful if a client has been moved or resized)
        self.state = PycaWM.idle
        self.start_positions = [0, 0]

    def key_press_reply(self, event):
        keysym_to_keycode = self.display.keysym_to_keycode
        if event.window == self.root:
            for keysym, mask in self.keys:
                if (event.detail == keysym_to_keycode(keysym) and
                    good_mask(event.state) == mask):
                    self.keys[(keysym, mask)](self)
                    return
        client = self.get_managed_client(event.window)
        if client is not None:
            for keysym, mask in Client.keys:
                if (event.detail == keysym_to_keycode(keysym) and
                    good_mask(event.state) == mask):
                    Client.keys[(keysym, mask)](client)

    def key_release_reply(self, event):
        pass

    @managed_client
    def property_reply(self, event, client):
        if event.atom == Xatom.WM_NAME:
            # update the client name
            client.name = event.window.get_wm_name()
            # redraw the window frame
            client.update_frame()

    @managed_client
    def motion_reply(self, event, client):
        def resize_at_edge(edge):
            if edge not in self.edge:
                return False
            elif edge == PycaWM.top:
                return (
                    (ny < 0 and pointer.root_y < client.geometry.y +
                     client.parent_border_width) or
                    (ny > 0 and pointer.root_y > client.geometry.y +
                     (client.parent_border_width * 2)))
            elif edge == PycaWM.bottom:
                return (
                    (ny < 0 and pointer.root_y < client.geometry.y +
                     client.geometry.height + client.parent_border_width) or
                    (ny > 0 and pointer.root_y > client.geometry.y +
                     client.geometry.height + (client.parent_border_width * 2)))
            elif edge == PycaWM.left:
                return (
                    (nx < 0 and pointer.root_x < client.geometry.x +
                     client.parent_border_width) or
                    (nx > 0 and pointer.root_x > client.geometry.x +
                     (client.parent_border_width * 2)))
            elif edge == PycaWM.right:
                return (
                    (nx < 0 and pointer.root_x < client.geometry.x +
                     client.geometry.width + client.parent_border_width) or
                    (nx > 0 and pointer.root_x > client.geometry.x +
                     client.geometry.width + (client.parent_border_width * 2)))
            else:
                return False

        pointer = self.root.query_pointer()
        if self.state == PycaWM.moving:
            client.move(dict(x=(client.geometry.x + pointer.root_x -
                                self.start_positions[0]),
                             y=(client.geometry.y + pointer.root_y -
                                self.start_positions[1])))
        elif self.state == PycaWM.resizing:
            nx = pointer.root_x - self.start_positions[0]
            ny = pointer.root_y - self.start_positions[1]

            client.resize(
                dict(left=(nx if resize_at_edge(PycaWM.left) else 0),
                     top=(ny if resize_at_edge(PycaWM.top) else 0),
                     right=(nx if resize_at_edge(PycaWM.right) else 0),
                     bottom=(ny if resize_at_edge(PycaWM.bottom) else 0)))

        self.start_positions = [pointer.root_x, pointer.root_y]

    def gravity_reply(self, event):
        pass

    @managed_client
    def enter_reply(self, event, client):
        if event.window == client.parent:
            pointer = self.root.query_pointer()
            x = pointer.root_x - client.geometry.x
            y = pointer.root_y - client.geometry.y
            edge = self.find_edge(client, x, y)

            client.parent.change_attributes(cursor=self.cursors[edge])
        elif event.window == client.title_bar:
            client.parent.change_attributes(cursor=self.cursors[PycaWM.idle,])

    @managed_client
    def leave_reply(self, event, client):
        client.parent.change_attributes(cursor=self.cursors[PycaWM.idle,])

    @managed_client
    def destroy_reply(self, event, client):
        client.destroy()
        self.set_new_focused_client()

    def circulate_request_reply(self, event):
        if event.place == X.PlaceOnTop:
            event.window.configure(stack_mode=X.Above)
        else: # event.place == X.PlaceOnBottom
            event.window.configure(stack_mode=X.Below)

    @managed_client
    def client_message_reply(self, event, client):
        # handle the Normal -> Iconic transition as described in the ICCCM
        if (event.data[0] == 32 and event.data[1][0] == Xutil.IconicState and
            event.client_type == atomsdb.WM_CHANGE_STATE and
            not client.is_hidden()):
            client.iconify()

    @managed_client
    def shape_reply(self, event, client):
        # the title bar has a shape too
        # but we want only to update the window shape
        if event.window == client.window:
            client.update_shape()

    def handle_events(self):
        with self.event_lock:
            self.display.flush()
            while self.display.pending_events():
                event = self.display.next_event()
                if event.type in self.dispatch:
                    try:
                        self.dispatch[event.type](event)
                    except error.XError, e:
                        print_exc(e)
                        print 'XError in PycaWM: ' + str(e)

    def event_loop(self):
        while 42:
            try:
                for disp in self.displays:
                    # flush the display event queue
                    if disp.pending_events() != 0:
                        disp.do_stuff()
                # displays have sockets too, take that into account
                ready_sockets, _, _ = select(self.displays + self.sockets,
                                             list(), list(), 1)
                for sock in ready_sockets:
                    sock.do_stuff()
            except (KeyboardInterrupt, Exception), e:
                print_exc(e)
                self.cleanup()
                raise SystemExit(0)

    def cleanup(self):
        for plugin in self.plugins:
            plugin.destroy()
        del self.plugins[:]
        try:
            self.root.ungrab_key(X.AnyKey, X.AnyModifier)
            self.unmanage_windows()
            self.free_colors()
            self.free_cursors()
            self.display.set_input_focus(X.PointerRoot, X.RevertToPointerRoot,
                                         X.CurrentTime)
            self.display.close()
        except error.ConnectionClosedError:
            pass

    def quit(self):
        self.cleanup()
        raise SystemExit(0)

    def switch_wm(self, new_wm, *args):
        self.cleanup()
        # clean the repl socket manually because the atexit hook
        # will never be called
        cleanup_unix_socket()
        os.execvp(new_wm, [new_wm] + list(args))
        print 'Cannot launch the wm %s' % str(new_wm)

    @classmethod
    def create_focuses(cls):
        def is_strict_focus_subclass(focus):
            if focus.__name__ == 'Focus':
                return False
            try:
                return issubclass(focus, focuses.Focus)
            except TypeError:
                return False
        gen_focuses = ((focus_name, focuses.__dict__.get(focus_name))
                       for focus_name in dir(focuses))
        cls.focuses = dict((focus_name, focus())
                           for focus_name, focus in gen_focuses
                           if not focus_name.startswith('_') and
                           is_strict_focus_subclass(focus))

    def set_focus(self, focus_name):
        def update_hooks(hook_fun, focus):
            for meth in set(self.dispatch.itervalues()):
                if hasattr(focus, meth.__name__):
                    hook_fun(meth, getattr(focus, meth.__name__))
        if focus_name in self.focuses:
            # remove old focus hooks
            if self.current_focus is not None:
                if self.current_focus in self.focuses:
                    focus = self.focuses[self.current_focus]
                    focus.destroy_focus(self)
                    update_hooks(remove_post_hook, focus)
                    self.current_focus = None
            # add new focus hooks
            focus = self.focuses[focus_name]
            focus.init_focus(self)
            update_hooks(add_post_hook, focus)
            self.current_focus = focus_name
        else:
            print 'focus not found! Keeping old focus policy'

    def give_focus(self, client):
        client.take_focus()

    def set_new_focused_client(self, force=True):
        if self.clients:
            client = self.clients[0]
            if (force or client.states.focused) and not client.is_hidden():
                # make sure the client is focused again
                client.states.focused = False
                self.give_focus(client)
                return

        self.root.set_input_focus(X.RevertToPointerRoot, X.CurrentTime)

    @property
    def stacked_clients(self):
        """Order: top client at the beginning of the list,
        bottom client at the end."""
        return filter(lambda client: client is not None,
                      reversed(map(self.get_managed_client,
                                   self.root.query_tree().children)))

    def move_pointer(self, window, x, y):
        if isinstance(window, Client):
            window = window.parent

        geo = window.get_geometry()
        ev_dict = dict(time=0,
                       root=self.root,
                       window=window,
                       same_screen=1,
                       child=X.NONE,
                       root_x=geo.x + x,
                       root_y=geo.y + y,
                       event_x=x,
                       event_y=y,
                       state=X.NONE,
                       detail=X.NotifyNormal)
        ev = protocol.event.MotionNotify(**ev_dict)
        window.send_event(event=ev, event_mask=X.PointerMotionMask)
        window.warp_pointer(x, y)

    def client_has_pointer(self, client):
        pointer = self.root.query_pointer()
        for client_ in self.stacked_clients:
            if client_.contain(pointer.root_x, pointer.root_y):
                return client == client_
        return False

    def client_at_point(self, point_x, point_y, clients=None):
        if clients is None:
            clients = self.stacked_clients
        found_client = None
        for client in clients:
            if client.is_hidden():
                continue
            if client.contain(point_x, point_y):
                found_client = client
                break
        return found_client

    @classmethod
    def create_placements(cls):
        def is_strict_placement_subclass(placement):
            if placement.__name__ == 'Placement':
                return False
            try:
                return issubclass(placement, placements.Placement)
            except TypeError:
                return False
        gen_placements = ((placement_name,
                           placements.__dict__.get(placement_name))
                          for placement_name in dir(placements))
        cls.placements = dict((placement_name, placement())
                              for placement_name, placement in gen_placements
                              if not placement_name.startswith('_') and
                              is_strict_placement_subclass(placement))

    def set_placements(self, placement_names):
        for placement_name in placement_names:
            if placement_name in self.placements:
                self.preferred_placements.append(placement_name)
            else:
                print 'placement not found! Cannot append it to the ' \
                    'preferred placements'

    def place_client(self, client):
        if client.wm_normal_hints is not None:
            # dialog box or windows with motif hints
            if ((client.states.transient or
                 client.motif_wm_hints is not None) and
                client.wm_normal_hints.flags & Xutil.PPosition):
                return

            # user defined position
            if client.wm_normal_hints.flags & Xutil.USPosition:
                return

        for placement_name in self.preferred_placements:
            if placement_name in self.placements:
                if self.placements[placement_name].place_client(self, client):
                    return

        print 'no good placement found! I put the client at (0, 0).'
        client.move(dict(x=0, y=0))

    def run_command(self, cmd, *args):
        pid = os.fork()
        # double fork for cleaning zombie processes
        # thanks to dwm for the idea
        if not pid:
            pid = os.fork()
            if not pid:
                os.setsid()
                try:
                    os.execvp(cmd, (cmd,) + args)
                except OSError:
                    # exec* failed?
                    os._exit(-1)
            os._exit(0)
        return os.wait()

    def find_client(self, wm_class):
        return [client for client in self.clients
                if (wm_class.strip().lower() ==
                    client.window.get_wm_class()[1].strip().lower())]
